from gym_softrobot.envs.octopus.flat_env import FlatEnv
from gym_softrobot.envs.octopus.arm_single_env import ArmSingleEnv
from gym_softrobot.envs.octopus.arm_two_env import ArmTwoEnv
from gym_softrobot.envs.octopus.arm_push_env import ArmPushEnv
from gym_softrobot.envs.octopus.arm_push_env import ArmPullWeightEnv
from gym_softrobot.envs.octopus.reach_env import ReachEnv
from gym_softrobot.envs.octopus.crawl_env import CrawlEnv
# from gym_softrobot.envs.octopus.crawlCurvature_env import CrawlCurvatureEnv
